==> flyte-0.0.1b2-py3-none-any.whl: flyte/_internal/imagebuild/docker_builder.py <==
import asyncio
import shutil
import subprocess
import tempfile
from pathlib import Path
from string import Template
from typing import ClassVar, Protocol, cast

import aiofiles
import click

from flyte._image import (
    AptPackages,
    Commands,
    CopyConfig,
    Env,
    Image,
    Layer,
    PipPackages,
    Requirements,
    UVProject,
    WorkDir,
    _DockerLines,
)
from flyte._logging import logger

_F_IMG_ID = "_F_IMG_ID"

UV_LOCK_INSTALL_TEMPLATE = Template("""\
WORKDIR /root
RUN --mount=type=cache,sharing=locked,mode=0777,target=/root/.cache/uv,id=uv \
    --mount=type=bind,target=uv.lock,src=uv.lock \
    --mount=type=bind,target=pyproject.toml,src=pyproject.toml \
    uv sync $PIP_INSTALL_ARGS
WORKDIR /

# Update PATH and UV_PYTHON to point to the venv created by uv sync
ENV PATH="/root/.venv/bin:$$PATH" \
    VIRTUALENV=/root/.venv \
    UV_PYTHON=/root/.venv/bin/python
""")

UV_PACKAGE_INSTALL_COMMAND_TEMPLATE = Template("""\
RUN --mount=type=cache,sharing=locked,mode=0777,target=/root/.cache/uv,id=uv \
    --mount=type=bind,target=requirements_uv.txt,src=requirements_uv.txt \
    uv pip install --prerelease=allow --python $$UV_PYTHON $PIP_INSTALL_ARGS
""")

APT_INSTALL_COMMAND_TEMPLATE = Template("""\
RUN --mount=type=cache,sharing=locked,mode=0777,target=/var/cache/apt,id=apt \
    apt-get update && apt-get install -y --no-install-recommends \
    $APT_PACKAGES
""")

UV_PYTHON_INSTALL_COMMAND = Template("""\
RUN --mount=type=cache,sharing=locked,mode=0777,target=/root/.cache/uv,id=uv \
    uv pip install --prerelease=allow $PIP_INSTALL_ARGS
""")

# uv pip install --python /root/env/bin/python
# new template
DOCKER_FILE_UV_BASE_TEMPLATE = Template("""\
#syntax=docker/dockerfile:1.5
FROM ghcr.io/astral-sh/uv:0.6.12 as uv
FROM $BASE_IMAGE

USER root

# Copy in uv so that later commands don't have to mount it in
COPY --from=uv /uv /usr/bin/uv

# Configure default envs
ENV UV_COMPILE_BYTECODE=1 \
    UV_LINK_MODE=copy \
    VIRTUALENV=/opt/venv \
    UV_PYTHON=/opt/venv/bin/python \
    PATH="/opt/venv/bin:$$PATH"

# Create a virtualenv with the user specified python version
RUN uv venv $$VIRTUALENV --python=$PYTHON_VERSION

# Adds nvidia just in case it exists
ENV PATH="$$PATH:/usr/local/nvidia/bin:/usr/local/cuda/bin" \
    LD_LIBRARY_PATH="/usr/local/nvidia/lib64:$$LD_LIBRARY_PATH"
""")

# This gets added on to the end of the dockerfile
DOCKER_FILE_BASE_FOOTER = Template("""\
ENV _F_IMG_ID=$F_IMG_ID
WORKDIR /root
SHELL ["/bin/bash", "-c"]
""")


class Handler(Protocol):
    @staticmethod
    async def handle(layer: Layer, context_path: Path, dockerfile: str) -> str: ...


class PipAndRequirementsHandler:
    @staticmethod
    async def handle(layer: PipPackages, context_path: Path, dockerfile: str) -> str:
        if isinstance(layer, Requirements):
            async with aiofiles.open(layer.file) as f:
                requirements = []
                async for line in f:
                    requirement = await line
                    requirements.append(requirement.strip())
        else:
            requirements = list(layer.packages) if layer.packages else []
        requirements_uv_path = context_path / "requirements_uv.txt"
        async with aiofiles.open(requirements_uv_path, "w") as f:
            reqs = "\n".join(requirements)
            await f.write(reqs)

        pip_install_args = []
        if layer.index_url:
            pip_install_args.append(f"--index-url {layer.index_url}")

        if layer.extra_index_urls:
            pip_install_args.extend([f"--extra-index-url {url}" for url in layer.extra_index_urls])

        if layer.pre:
            pip_install_args.append("--pre")

        if layer.extra_args:
            pip_install_args.append(layer.extra_args)

        pip_install_args.extend(["--requirement", "requirements_uv.txt"])

        delta = UV_PACKAGE_INSTALL_COMMAND_TEMPLATE.substitute(PIP_INSTALL_ARGS=" ".join(pip_install_args))
        dockerfile += delta

        return dockerfile


class _DockerLinesHandler:
    @staticmethod
    async def handle(layer: _DockerLines, context_path: Path, dockerfile: str) -> str:
        # Add the lines to the dockerfile
        for line in layer.lines:
            dockerfile += f"\n{line}\n"

        return dockerfile


class EnvHandler:
    @staticmethod
    async def handle(layer: Env, context_path: Path, dockerfile: str) -> str:
        # Add the env vars to the dockerfile
        for key, value in layer.env_vars:
            dockerfile += f"\nENV {key}={value}\n"

        return dockerfile


class AptPackagesHandler:
    @staticmethod
    async def handle(layer: AptPackages, context_path: Path, dockerfile: str) -> str:
        packages = layer.packages
        delta = APT_INSTALL_COMMAND_TEMPLATE.substitute(APT_PACKAGES=" ".join(packages))
        dockerfile += delta

        return dockerfile


class UVProjectHandler:
    @staticmethod
    async def handle(layer: UVProject, context_path: Path, dockerfile: str) -> str:
        # copy the two files
        shutil.copy(layer.pyproject, context_path)
        shutil.copy(layer.uvlock, context_path)

        # --locked: Assert that the `uv.lock` will remain unchanged
        # --no-dev: Omit the development dependency group
        # --no-install-project: Do not install the current project
        additional_pip_install_args = ["--locked", "--no-dev", "--no-install-project"]
        delta = UV_LOCK_INSTALL_TEMPLATE.substitute(PIP_INSTALL_ARGS=" ".join(additional_pip_install_args))
        dockerfile += delta

        return dockerfile


class CopyConfigHandler:
    @staticmethod
    async def handle(layer: CopyConfig, context_path: Path, dockerfile: str) -> str:
        # Copy the source config file or directory to the context path
        abs_path = layer.context_source.absolute()
        dest_path = context_path / abs_path.name
        image_dest_path = layer.image_dest + "/" + abs_path.name
        if layer.context_source.is_file():
            # Copy the file
            shutil.copy(abs_path, dest_path)
        elif layer.context_source.is_dir():
            # Copy the entire directory
            shutil.copytree(abs_path, dest_path)
        else:
            raise ValueError(f"Source path is neither file nor directory: {layer.context_source}")

        # Add a copy command to the dockerfile
        dockerfile += f"\nCOPY {abs_path.name} {image_dest_path}\n"

        return dockerfile


class CommandsHandler:
    @staticmethod
    async def handle(layer: Commands, context_path: Path, dockerfile: str) -> str:
        # Append raw commands to the dockerfile
        for command in layer.commands:
            dockerfile += f"\nRUN {command}\n"

        return dockerfile


class WorkDirHandler:
    @staticmethod
    async def handle(layer: WorkDir, context_path: Path, dockerfile: str) -> str:
        # cd to the workdir
        dockerfile += f"\nWORKDIR {layer.workdir}\n"

        return dockerfile


async def _process_layer(layer: Layer, context_path: Path, dockerfile: str) -> str:
    match layer:
        case Requirements() | PipPackages():
            # Handle pip packages and requirements
            dockerfile = await PipAndRequirementsHandler.handle(layer, context_path, dockerfile)

        case AptPackages():
            # Handle apt packages
            dockerfile = await AptPackagesHandler.handle(layer, context_path, dockerfile)

        case UVProject():
            # Handle UV project
            dockerfile = await UVProjectHandler.handle(layer, context_path, dockerfile)

        case CopyConfig():
            # Handle local files and folders
            dockerfile = await CopyConfigHandler.handle(layer, context_path, dockerfile)

        case Commands():
            # Handle commands
            dockerfile = await CommandsHandler.handle(layer, context_path, dockerfile)

        case WorkDir():
            # Handle workdir
            dockerfile = await WorkDirHandler.handle(layer, context_path, dockerfile)

        case Env():
            # Handle environment variables
            dockerfile = await EnvHandler.handle(layer, context_path, dockerfile)

        case _DockerLines():
            # Only for internal use
            dockerfile = await _DockerLinesHandler.handle(layer, context_path, dockerfile)

        case _:
            raise NotImplementedError(f"Layer type {type(layer)} not supported")

    return dockerfile


class DockerImageBuilder:
    """Image builder using Docker and buildkit."""

    builder_type: ClassVar = "docker"
    _builder_name: ClassVar = "flytex"

    async def build_image(self, image: Image, dry_run: bool = False) -> str:
        if image.is_final:
            if image._layers:
                raise ValueError("Image is a default image and should already be built")

        if image.dockerfile:
            # If a dockerfile is provided, use it directly
            return await self._build_from_dockerfile(image, push=True)

        return await self._build_image(
            image,
            push=True,
            dry_run=dry_run,
        )

    async def _build_from_dockerfile(self, image: Image, push: bool) -> str:
        """
        Build the image from a provided Dockerfile.
        """
        command = [
            "docker",
            "build",
            "--tag",
            f"{image.uri}",
            "--platform",
            ",".join(image.platform),
            ".",
        ]

        if image.registry and push:
            command.append("--push")

        concat_command = " ".join(command)
        logger.debug(f"Build command: {concat_command}")
        click.secho(f"Run command: {concat_command} ", fg="blue")

        await asyncio.to_thread(subprocess.run, command, cwd=str(cast(Path, image.dockerfile).cwd()), check=True)

        return image.uri

    @staticmethod
    async def _ensure_buildx_builder():
        """Ensure there is a docker buildx builder called flyte"""
        # Check if buildx is available
        try:
            await asyncio.to_thread(
                subprocess.run, ["docker", "buildx", "version"], check=True, stdout=subprocess.DEVNULL
            )
        except subprocess.CalledProcessError:
            raise RuntimeError("Docker buildx is not available. Make sure BuildKit is installed and enabled.")

        # List builders
        result = await asyncio.to_thread(
            subprocess.run, ["docker", "buildx", "ls"], capture_output=True, text=True, check=True
        )
        builders = result.stdout

        # Check if there's any usable builder
        if DockerImageBuilder._builder_name not in builders:
            # No default builder found, create one
            logger.info("No buildx builder found, creating one...")
            await asyncio.to_thread(
                subprocess.run,
                [
                    "docker",
                    "buildx",
                    "create",
                    "--name",
                    DockerImageBuilder._builder_name,
                    "--platform",
                    "linux/amd64,linux/arm64",
                ],
                check=True,
            )
        else:
            logger.info("Buildx builder already exists.")

    async def _build_image(self, image: Image, *, push: bool = True, dry_run: bool = False) -> str:
        """
        if default image (only base image and locked), raise an error, don't have a dockerfile
        if dockerfile, just build
        in the main case, get the default Dockerfile template
          - start from the base image
          - use python to create a default venv and export variables

          Then for the layers
          - for each layer
            - find the appropriate layer handler
            - call layer handler with the context dir and the dockerfile
              - handler can choose to do something (copy files from local) to the context and update the dockerfile
                contents, returning the new string
        """
        # For testing, set `push=False` to just build the image locally and not push to
        # registry.

        await DockerImageBuilder._ensure_buildx_builder()

        with tempfile.TemporaryDirectory() as tmp_dir:
            logger.warning(f"Temporary directory: {tmp_dir}")
            tmp_path = Path(tmp_dir)

            dockerfile = DOCKER_FILE_UV_BASE_TEMPLATE.substitute(
                BASE_IMAGE=image.base_image,
                PYTHON_VERSION=f"{image.python_version[0]}.{image.python_version[1]}",
            )

            for layer in image._layers:
                dockerfile = await _process_layer(layer, tmp_path, dockerfile)

            dockerfile += DOCKER_FILE_BASE_FOOTER.substitute(F_IMG_ID=image.uri)

            dockerfile_path = tmp_path / "Dockerfile"
            async with aiofiles.open(dockerfile_path, mode="w") as f:
                await f.write(dockerfile)

            command = [
                "docker",
                "buildx",
                "build",
                "--builder",
                DockerImageBuilder._builder_name,
                "--tag",
                f"{image.uri}",
                "--platform",
                ",".join(image.platform),
                "--push" if push else "--load",
            ]

            if image.registry and push:
                command.append("--push")
            command.append(tmp_dir)

            concat_command = " ".join(command)
            logger.debug(f"Build command: {concat_command}")
            if dry_run:
                click.secho("Dry run for docker builder...")
                click.secho(f"Context path: {tmp_path}")
                click.secho(f"Dockerfile: {dockerfile}")
                click.secho(f"Command: {concat_command}")
                return image.uri
            else:
                click.secho(f"Run command: {concat_command} ", fg="blue")

            await asyncio.to_thread(subprocess.run, command, check=True)

            return image.uri

==> flyte-0.2.0b20-py3-none-any.whl: flyte/_internal/imagebuild/docker_builder.py <==
import asyncio
import os
import shutil
import subprocess
import tempfile
import typing
from pathlib import Path
from string import Template
from typing import ClassVar, Optional, Protocol, cast

import aiofiles
import click

from flyte._image import (
    AptPackages,
    Commands,
    CopyConfig,
    Env,
    Image,
    Layer,
    PipPackages,
    PythonWheels,
    Requirements,
    UVProject,
    WorkDir,
    _DockerLines,
)
from flyte._internal.imagebuild.image_builder import (
    DockerAPIImageChecker,
    ImageBuilder,
    ImageChecker,
    LocalDockerCommandImageChecker,
    LocalPodmanCommandImageChecker,
)
from flyte._logging import logger

_F_IMG_ID = "_F_IMG_ID"
FLYTE_DOCKER_BUILDER_CACHE_FROM = "FLYTE_DOCKER_BUILDER_CACHE_FROM"
FLYTE_DOCKER_BUILDER_CACHE_TO = "FLYTE_DOCKER_BUILDER_CACHE_TO"

UV_LOCK_INSTALL_TEMPLATE = Template("""\
WORKDIR /root
RUN --mount=type=cache,sharing=locked,mode=0777,target=/root/.cache/uv,id=uv \
    --mount=type=bind,target=uv.lock,src=uv.lock \
    --mount=type=bind,target=pyproject.toml,src=pyproject.toml \
    uv sync $PIP_INSTALL_ARGS
WORKDIR /

# Update PATH and UV_PYTHON to point to the venv created by uv sync
ENV PATH="/root/.venv/bin:$$PATH" \
    VIRTUALENV=/root/.venv \
    UV_PYTHON=/root/.venv/bin/python
""")

UV_PACKAGE_INSTALL_COMMAND_TEMPLATE = Template("""\
RUN --mount=type=cache,sharing=locked,mode=0777,target=/root/.cache/uv,id=uv \
    --mount=type=bind,target=requirements_uv.txt,src=requirements_uv.txt \
    uv pip install --prerelease=allow --python $$UV_PYTHON $PIP_INSTALL_ARGS
""")

UV_WHEEL_INSTALL_COMMAND_TEMPLATE = Template("""\
RUN --mount=type=cache,sharing=locked,mode=0777,target=/root/.cache/uv,id=wheel \
    --mount=source=/dist,target=/dist,type=bind \
    uv pip install --prerelease=allow --python $$UV_PYTHON $PIP_INSTALL_ARGS
""")

APT_INSTALL_COMMAND_TEMPLATE = Template("""\
RUN --mount=type=cache,sharing=locked,mode=0777,target=/var/cache/apt,id=apt \
    apt-get update && apt-get install -y --no-install-recommends \
    $APT_PACKAGES
""")

UV_PYTHON_INSTALL_COMMAND = Template("""\
RUN --mount=type=cache,sharing=locked,mode=0777,target=/root/.cache/uv,id=uv \
    uv pip install --prerelease=allow $PIP_INSTALL_ARGS
""")

# uv pip install --python /root/env/bin/python
# new template
DOCKER_FILE_UV_BASE_TEMPLATE = Template("""\
#syntax=docker/dockerfile:1.5
FROM ghcr.io/astral-sh/uv:0.6.12 as uv
FROM $BASE_IMAGE

USER root

# Copy in uv so that later commands don't have to mount it in
COPY --from=uv /uv /usr/bin/uv

# Configure default envs
ENV UV_COMPILE_BYTECODE=1 \
    UV_LINK_MODE=copy \
    VIRTUALENV=/opt/venv \
    UV_PYTHON=/opt/venv/bin/python \
    PATH="/opt/venv/bin:$$PATH"

# Create a virtualenv with the user specified python version
RUN uv venv $$VIRTUALENV --python=$PYTHON_VERSION

# Adds nvidia just in case it exists
ENV PATH="$$PATH:/usr/local/nvidia/bin:/usr/local/cuda/bin" \
    LD_LIBRARY_PATH="/usr/local/nvidia/lib64:$$LD_LIBRARY_PATH"
""")

# This gets added on to the end of the dockerfile
DOCKER_FILE_BASE_FOOTER = Template("""\
ENV _F_IMG_ID=$F_IMG_ID
WORKDIR /root
SHELL ["/bin/bash", "-c"]
""")


class Handler(Protocol):
    @staticmethod
    async def handle(layer: Layer, context_path: Path, dockerfile: str) -> str: ...


class PipAndRequirementsHandler:
    @staticmethod
    async def handle(layer: PipPackages, context_path: Path, dockerfile: str) -> str:
        if isinstance(layer, Requirements):
            if not layer.file.exists():
                raise FileNotFoundError(f"Requirements file {layer.file} does not exist")
            if not layer.file.is_file():
                raise ValueError(f"Requirements file {layer.file} is not a file")

            async with aiofiles.open(layer.file) as f:
                requirements = []
                async for line in f:
                    requirement = line
                    requirements.append(requirement.strip())
        else:
            requirements = list(layer.packages) if layer.packages else []
        requirements_uv_path = context_path / "requirements_uv.txt"
        async with aiofiles.open(requirements_uv_path, "w") as f:
            reqs = "\n".join(requirements)
            await f.write(reqs)

        pip_install_args = layer.get_pip_install_args()
        pip_install_args.extend(["--requirement", "requirements_uv.txt"])

        delta = UV_PACKAGE_INSTALL_COMMAND_TEMPLATE.substitute(PIP_INSTALL_ARGS=" ".join(pip_install_args))
        dockerfile += delta

        return dockerfile


class PythonWheelHandler:
    @staticmethod
    async def handle(layer: PythonWheels, context_path: Path, dockerfile: str) -> str:
        shutil.copytree(layer.wheel_dir, context_path / "dist", dirs_exist_ok=True)
        pip_install_args = layer.get_pip_install_args()
        pip_install_args.extend(["/dist/*.whl"])

        delta = UV_WHEEL_INSTALL_COMMAND_TEMPLATE.substitute(PIP_INSTALL_ARGS=" ".join(pip_install_args))
        dockerfile += delta

        return dockerfile


class _DockerLinesHandler:
    @staticmethod
    async def handle(layer: _DockerLines, context_path: Path, dockerfile: str) -> str:
        # Add the lines to the dockerfile
        for line in layer.lines:
            dockerfile += f"\n{line}\n"

        return dockerfile


class EnvHandler:
    @staticmethod
    async def handle(layer: Env, context_path: Path, dockerfile: str) -> str:
        # Add the env vars to the dockerfile
        for key, value in layer.env_vars:
            dockerfile += f"\nENV {key}={value}\n"

        return dockerfile


class AptPackagesHandler:
    @staticmethod
    async def handle(layer: AptPackages, context_path: Path, dockerfile: str) -> str:
        packages = layer.packages
        delta = APT_INSTALL_COMMAND_TEMPLATE.substitute(APT_PACKAGES=" ".join(packages))
        dockerfile += delta

        return dockerfile


class UVProjectHandler:
    @staticmethod
    async def handle(layer: UVProject, context_path: Path, dockerfile: str) -> str:
        # copy the two files
        shutil.copy(layer.pyproject, context_path)
        shutil.copy(layer.uvlock, context_path)

        # --locked: Assert that the `uv.lock` will remain unchanged
        # --no-dev: Omit the development dependency group
        # --no-install-project: Do not install the current project
        additional_pip_install_args = ["--locked", "--no-dev", "--no-install-project"]
        delta = UV_LOCK_INSTALL_TEMPLATE.substitute(PIP_INSTALL_ARGS=" ".join(additional_pip_install_args))
        dockerfile += delta

        return dockerfile


class CopyConfigHandler:
    @staticmethod
    async def handle(layer: CopyConfig, context_path: Path, dockerfile: str) -> str:
        # Copy the source config file or directory to the context path
        if layer.src.is_absolute() or ".." in str(layer.src):
            dst_path = context_path / str(layer.src.absolute()).replace("/", "./_flyte_abs_context/")
        else:
            dst_path = context_path / layer.src

        dst_path.parent.mkdir(parents=True, exist_ok=True)
        abs_path = layer.src.absolute()
        if layer.src.is_file():
            # Copy the file
            shutil.copy(abs_path, dst_path)
        elif layer.src.is_dir():
            # Copy the entire directory
            shutil.copytree(abs_path, dst_path, dirs_exist_ok=True)
        else:
            raise ValueError(f"Source path is neither file nor directory: {layer.src}")

        # Add a copy command to the dockerfile
        dockerfile += f"\nCOPY {dst_path.relative_to(context_path)} {layer.dst}\n"

        return dockerfile


class CommandsHandler:
    @staticmethod
    async def handle(layer: Commands, _: Path, dockerfile: str) -> str:
        # Append raw commands to the dockerfile
        for command in layer.commands:
            dockerfile += f"\nRUN {command}\n"

        return dockerfile


class WorkDirHandler:
    @staticmethod
    async def handle(layer: WorkDir, context_path: Path, dockerfile: str) -> str:
        # cd to the workdir
        dockerfile += f"\nWORKDIR {layer.workdir}\n"

        return dockerfile


async def _process_layer(layer: Layer, context_path: Path, dockerfile: str) -> str:
    match layer:
        case PythonWheels():
            # Handle Python wheels
            dockerfile = await PythonWheelHandler.handle(layer, context_path, dockerfile)

        case Requirements() | PipPackages():
            # Handle pip packages and requirements
            dockerfile = await PipAndRequirementsHandler.handle(layer, context_path, dockerfile)

        case AptPackages():
            # Handle apt packages
            dockerfile = await AptPackagesHandler.handle(layer, context_path, dockerfile)

        case UVProject():
            # Handle UV project
            dockerfile = await UVProjectHandler.handle(layer, context_path, dockerfile)

        case CopyConfig():
            # Handle local files and folders
            dockerfile = await CopyConfigHandler.handle(layer, context_path, dockerfile)

        case Commands():
            # Handle commands
            dockerfile = await CommandsHandler.handle(layer, context_path, dockerfile)

        case WorkDir():
            # Handle workdir
            dockerfile = await WorkDirHandler.handle(layer, context_path, dockerfile)

        case Env():
            # Handle environment variables
            dockerfile = await EnvHandler.handle(layer, context_path, dockerfile)

        case _DockerLines():
            # Only for internal use
            dockerfile = await _DockerLinesHandler.handle(layer, context_path, dockerfile)

        case _:
            raise NotImplementedError(f"Layer type {type(layer)} not supported")

    return dockerfile


class DockerImageBuilder(ImageBuilder):
    """Image builder using Docker and buildkit."""

    builder_type: ClassVar = "docker"
    _builder_name: ClassVar = "flytex"

    def get_checkers(self) -> Optional[typing.List[typing.Type[ImageChecker]]]:
        # Can get a public token for docker.io but ghcr requires a pat, so harder to get the manifest anonymously
        return [LocalDockerCommandImageChecker, LocalPodmanCommandImageChecker, DockerAPIImageChecker]

    async def build_image(self, image: Image, dry_run: bool = False) -> str:
        if len(image._layers) == 0:
            logger.warning("No layers to build, returning the image URI as is.")
            return image.uri

        if image.dockerfile:
            # If a dockerfile is provided, use it directly
            return await self._build_from_dockerfile(image, push=True)

        return await self._build_image(
            image,
            push=True,
            dry_run=dry_run,
        )

    async def _build_from_dockerfile(self, image: Image, push: bool) -> str:
        """
        Build the image from a provided Dockerfile.
        """
        command = [
            "docker",
            "build",
            "--tag",
            f"{image.uri}",
            "--platform",
            ",".join(image.platform),
            ".",
        ]

        if image.registry and push:
            command.append("--push")

        concat_command = " ".join(command)
        logger.debug(f"Build command: {concat_command}")
        click.secho(f"Run command: {concat_command} ", fg="blue")

        await asyncio.to_thread(subprocess.run, command, cwd=str(cast(Path, image.dockerfile).cwd()), check=True)

        return image.uri

    @staticmethod
    async def _ensure_buildx_builder():
        """Ensure there is a docker buildx builder called flyte"""
        # Check if buildx is available
        try:
            await asyncio.to_thread(
                subprocess.run, ["docker", "buildx", "version"], check=True, stdout=subprocess.DEVNULL
            )
        except subprocess.CalledProcessError:
            raise RuntimeError("Docker buildx is not available. Make sure BuildKit is installed and enabled.")

        # List builders
        result = await asyncio.to_thread(
            subprocess.run, ["docker", "buildx", "ls"], capture_output=True, text=True, check=True
        )
        builders = result.stdout

        # Check if there's any usable builder
        if DockerImageBuilder._builder_name not in builders:
            # No default builder found, create one
            logger.info("No buildx builder found, creating one...")
            await asyncio.to_thread(
                subprocess.run,
                [
                    "docker",
                    "buildx",
                    "create",
                    "--name",
                    DockerImageBuilder._builder_name,
                    "--platform",
                    "linux/amd64,linux/arm64",
                ],
                check=True,
            )
        else:
            logger.info("Buildx builder already exists.")

    async def _build_image(self, image: Image, *, push: bool = True, dry_run: bool = False) -> str:
        """
        if default image (only base image and locked), raise an error, don't have a dockerfile
        if dockerfile, just build
        in the main case, get the default Dockerfile template
          - start from the base image
          - use python to create a default venv and export variables

          Then for the layers
          - for each layer
            - find the appropriate layer handler
            - call layer handler with the context dir and the dockerfile
              - handler can choose to do something (copy files from local) to the context and update the dockerfile
                contents, returning the new string
        """
        # For testing, set `push=False` to just build the image locally and not push to
        # registry.

        await DockerImageBuilder._ensure_buildx_builder()

        with tempfile.TemporaryDirectory() as tmp_dir:
            logger.warning(f"Temporary directory: {tmp_dir}")
            tmp_path = Path(tmp_dir)

            dockerfile = DOCKER_FILE_UV_BASE_TEMPLATE.substitute(
                BASE_IMAGE=image.base_image,
                PYTHON_VERSION=f"{image.python_version[0]}.{image.python_version[1]}",
            )

            for layer in image._layers:
                dockerfile = await _process_layer(layer, tmp_path, dockerfile)

            dockerfile += DOCKER_FILE_BASE_FOOTER.substitute(F_IMG_ID=image.uri)

            dockerfile_path = tmp_path / "Dockerfile"
            async with aiofiles.open(dockerfile_path, mode="w") as f:
                await f.write(dockerfile)

            command = [
                "docker",
                "buildx",
                "build",
                "--builder",
                DockerImageBuilder._builder_name,
                "--tag",
                f"{image.uri}",
                "--platform",
                ",".join(image.platform),
            ]

            cache_from = os.getenv(FLYTE_DOCKER_BUILDER_CACHE_FROM)
            cache_to = os.getenv(FLYTE_DOCKER_BUILDER_CACHE_TO)
            if cache_from and cache_to:
                command[3:3] = [
                    f"--cache-from={cache_from}",
                    f"--cache-to={cache_to}",
                ]

            if image.registry and push:
                command.append("--push")
            else:
                command.append("--load")
            command.append(tmp_dir)

            concat_command = " ".join(command)
            logger.debug(f"Build command: {concat_command}")
            if dry_run:
                click.secho("Dry run for docker builder...")
                click.secho(f"Context path: {tmp_path}")
                click.secho(f"Dockerfile: {dockerfile}")
                click.secho(f"Command: {concat_command}")
                return image.uri
            else:
                click.secho(f"Run command: {concat_command} ", fg="blue")

            await asyncio.to_thread(subprocess.run, command, check=True)

            return image.uri
